' //////////////////////////////////////////////////////////////////////
' TITLE: EPM_Defnder_Sound_Engine1_xxx.spin
'
' DESCRIPTION:
'
' AUTHOR:  Eric Moyer, Copyright (C)2007
'
'
'  Rev  Date      Description
'  ---  --------  ---------------------------------------------------
'  001  11-17-07  Initial Creation
'  003  11-18-07  First working audio.
'  006  11-18-07  Increase edge rate on square waves.       
'  009  11-18-07  Add multi sound support.  Add play and stop controls.
'  012  11-19-07  Move to sound mixing model and change edge emphasis implementation
'  013  11-27-07  Experiment with mixed dual triangle wave for thrust
'  014  11-27-07  Go back to single triangle for thrust and refine parameters a little.
'  016  11-27-07  Add start games sound
'  017  11-27-07  Add YouDie sound (large explosion)
'  018  11-27-07  Add Warp and Mutant
'  019  11-27-07  Add LanderFire
'  020  11-28-07  Test random LFSR generator
'  021  11-28-07  Modify Thrust to use LFSR.  Temporarily disable LanderFire for code space
'  022  11-28-07  Optimized code space a little.
'                 Delete unused constants and vars.
'                 Added back in LanderFire.
'                 Used LFSR for all random generation.  
'  025  11-29-07  Merge all explosion sounds (Fire, YouDie, Warp, HumanDie, PlanetExplode) into
'                 same core handler.
'  027  12-01-07  Split sound into two engines.
'                 Moved LevelDone and StartGame into engine 2.
'                 Added Human Abducted
'                 Added capability to stop any sound (previously only applied to thrust)
'                 Added Human Caught
'
' //////////////////////////////////////////////////////////////////////
'
'  Audio channel usage:
'     0: Thrust
'     1: (All others)
'
' //////////////////////////////////////////////////////////////////////
CON

  CLOCK_RATE_HYDRA       = 80_000_000
  CLOCK_RATE_HYBRID      = 96_000_000
  PIN__AUDIO             = %00000000_00000000_00000000_10000000

  AUDIO_SAMPLE_RATE      = 22_000  'Audio generation rate

  '------------------------------------
  'HARDWARE IDs
  '------------------------------------
  HARDWARE_HYDRA         = 1 
  HARDWARE_HYBRID        = 2
  
  '------------------------------------
  'SOUND EFFECT IDs
  '------------------------------------

  'Internally generated IDs (do not use in START calls)
  SOUND__INT_EXPLODE     = %00000000_00000000_00000000_00000001 'Handles all "explosion" types
  SOUND__INT_IDLE_THRUST = %00000000_00000000_00000000_00000010 'Handles idle thrust sound

  'Internally masks (do not use in START calls)  
  MASK_EXPLOSIONS        = %00000000_00000000_00000000_01111100 'All explostion types
  MASK_UNSUPPORTED       = %00000000_00001110_00000000_00000000 'Sounds unsupported by this engine (i.e. sounds
                                                                ' handled by the other engine)
  'Exteranally accessible IDs
  SOUND__HUMAN_DIE       = %00000000_00000000_00000000_00000100 
  SOUND__FIRE            = %00000000_00000000_00000000_00001000 
  SOUND__WARP            = %00000000_00000000_00000000_00010000 
  SOUND__YOU_DIE         = %00000000_00000000_00000000_00100000 
  SOUND__PLANET_EXPLODE  = %00000000_00000000_00000000_01000000 

  SOUND__THRUST          = %00000000_00000000_00000000_10000000 
  SOUND__MUTANT          = %00000000_00000000_00000001_00000000 
  SOUND__LANDER_FIRE     = %00000000_00000000_00000010_00000000 
  SOUND__LANDER_DIE      = %00000000_00000000_00000100_00000000 
  SOUND__HUMAN_ABDUCTED  = %00000000_00000000_00001000_00000000 
  SOUND__HUMAN_FALLING   = %00000000_00000000_00010000_00000000 
  SOUND__HUMAN_CAUGHT    = %00000000_00000000_00100000_00000000 
  SOUND__HUMAN_RETURNED  = %00000000_00000000_01000000_00000000 
  SOUND__SWARMER         = %00000000_00000000_10000000_00000000 
  SOUND__EXTRA_LIFE      = %00000000_00000001_00000000_00000000 
  SOUND__LEVEL_DONE      = %00000000_00000010_00000000_00000000 
  SOUND__START_GAME      = %00000000_00000100_00000000_00000000 
  SOUND__HIGHSCORE       = %00000000_00001000_00000000_00000000 
  
  '------------------------------------
  'Flags
  '------------------------------------  
  FLAG_RISING_EDGE       = %00000001
  FLAG_FALLING_EDGE      = %00000010

  '------------------------------------
  'Misc
  '------------------------------------  
  MUTANT_SAMPLE_WIDTH_INIT = 2
  MUTANT_SAMPLES_PER_LWORD = 16

  LFIRE_SAMPLE_WIDTH_INIT = 3 
  LFIRE_SAMPLES_PER_LWORD = 32

  COG_STOPPED             = 99
  
OBJ
  'dbg       :       "PasDebug"                       '<---- Add for Debugger
  
VAR
   long               this_cog 

PUB init 
   this_cog := COG_STOPPED
  
PUB start (sfx_control_block_p)
  if this_cog == COG_STOPPED
    this_cog := cognew(@sfx_entry,sfx_control_block_p)
'   dbg.start(31,30,@sfx_entry)                 '<---- Add for Debugger   

PUB stop 
  if (this_cog <> COG_STOPPED)  
    cogstop(this_cog)
    this_cog := COG_STOPPED

'///////////////////////////////////////////////////////////////////////
' DATA SECTION /////////////////////////////////////////////////////////
'///////////////////////////////////////////////////////////////////////

DAT
                        
'------------------------------------
'Entry
'------------------------------------
                        org        0
                        
sfx_entry

'  --------- Debugger Kernel add this at Entry (Addr 0) ---------
'   long $34FC1202,$6CE81201,$83C120B,$8BC0E0A,$E87C0E03,$8BC0E0A
'   long $EC7C0E05,$A0BC1207,$5C7C0003,$5C7C0003,$7FFC,$7FF8
'  -------------------------------------------------------------- 

                        'get pointers to parameters in the Effect Control block
                        mov     r1,PAR             'Get pointer to Effect Control block

                        'Get Pointer to sfx_start_flags
                        mov     p_sfx_start_flags, r1
                        add     r1, #4

                        'Get Pointer to sfx_stop_flags
                        mov     p_sfx_stop_flags, r1
                        add     r1, #4

                        'Get (and process) sfx_hardware_type
                        rdlong  r2, r1             ' Get hardware id
                        cmp     r2, #HARDWARE_HYDRA wz
              if_z      mov     sample_interval_ticks, SAMPLE_INTERVAL_HYDRA
              if_nz     mov     sample_interval_ticks, SAMPLE_INTERVAL_HYBRID  

'------------------------------------
'Effect processing loop
'------------------------------------

                        '------------------------------------
                        'Init 
                        '------------------------------------
                        mov     sfx_playing_mask, #0

                        'Configure audio pin
                        andn    outa,_PIN__AUDIO       ' Set audio pin LOW          "0"
                        or      dira,_PIN__AUDIO       ' Set audio pin to OUTPUT
                        mov     ctra, CTRAVAL
                        mov     frqa, LONG_HALF
                        mov     phsa, #$ff

                        'DEBUG LED
                        'or      dira, #1

                        'seed the LFSR pseudo-random generator
                        mov     random_number, #1  

                        mov     start_request_previous, #0
                        mov     start_request_current, #0
                        mov     start_request_actions, #0
                        mov     sfx_in_progress, #0
:restart                        
                        mov     next_sample_start_time, cnt
                        add     next_sample_start_time, sample_interval_ticks

                        mov     previous_pulse_level, #0


:sample_sync            'wait for next audio sample start time
                        waitcnt next_sample_start_time, sample_interval_ticks

                        
                        mov     flags, #0
                        mov     AUDIO_channel_0, #0
                        mov     AUDIO_channel_1, #0

                        '------------------------------------
                        'Generate LFSR Pseudo random number                                      
                        '------------------------------------
                        mov    x, #32                              'Generate 32 bits
:lfsr_loop              mov    y, random_number
                        and    y, #1
                        neg    y, y
                        and    y, LFSR_32_TAPS
          
                        xor    random_number, y
                        ror    random_number, #1
                        sub    x, #1  wz
              if_nz     jmp    #:lfsr_loop  
 
                        
                        '------------------------------------
                        'Process SFX start requests
                        '------------------------------------
                        'Determine if any SFX's have been requested to start
                        rdlong  start_request_current, p_sfx_start_flags       'get current start flags
                        mov     start_request_actions, start_request_current   'valid starts are any newly set bits
                        andn    start_request_actions, start_request_previous
                        mov     start_request_previous, start_request_current 'save current as previoius

                        '------------------------------------
                        'Process SFX stop requests
                        '------------------------------------
                        'Determine if any SFX's have been requested to stop
                        rdlong  stop_request_current, p_sfx_stop_flags         'get current stop flags
                        andn    sfx_in_progress, stop_request_current

                        '------------------------------------
                        'START: <All explostion types>
                        '------------------------------------
                        '
                        '  Common init code for all explosion types
                        '
                        '------------------------------------
                        test    start_request_actions, #MASK_EXPLOSIONS   wz
             if_nz      mov     EXPLODE_phase_ticks, FIRE_PHASE_TICK_INIT
             if_nz      mov     EXPLODE_chunk_width, #1
             if_nz      mov     EXPLODE_chunk_step,  #1   
             if_nz      mov     EXPLODE_pulse_ticks, #1
             if_nz      mov     EXPLODE_pulse_sense, #1
             if_nz      mov     EXPLODE_pulse_level, LONG_HALF
             if_nz      mov     EXPLODE_state_counter, #0
             if_nz      mov     random_number, #7                'Re-seed the random number generator for all explosion starts
                                       
                        '------------------------------------
                        'START: Human Die
                        '------------------------------------
                        mov     sfx_id_mask,#SOUND__HUMAN_DIE
                        test    start_request_actions, sfx_id_mask    wz

                        'Configure sound
             if_nz      mov     EXPLODE_chunk_width, #$4
             if_nz      mov     EXPLODE_chunk_step,  #4 
             if_nz      mov     EXPLODE_type, sfx_id_mask

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, #SOUND__INT_EXPLODE  

                        '------------------------------------
                        'START: Fire 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

                        'Configure sound
             if_nz      mov     EXPLODE_type, sfx_id_mask
             if_nz      shr     EXPLODE_pulse_level, #3

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, #SOUND__INT_EXPLODE

                        '------------------------------------
                        'START: Warp 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

                        'Configure sound
             if_nz      mov     EXPLODE_phase_ticks, FIRE_PHASE_TICK_INIT
             if_nz      mov     EXPLODE_chunk_width, #40
             if_nz      neg     EXPLODE_chunk_step,  #1            'step = -1
             if_nz      mov     EXPLODE_type, sfx_id_mask
             if_nz      shr     EXPLODE_pulse_level, #3    

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, #SOUND__INT_EXPLODE

                        '------------------------------------
                        'START: You Die 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

                        'Configure sound
             if_nz      mov     EXPLODE_type, sfx_id_mask

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, #SOUND__INT_EXPLODE

                        '------------------------------------
                        'START: Planet Explode
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

                        'Configure sound
             if_nz      mov     EXPLODE_type, sfx_id_mask     

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, #SOUND__INT_EXPLODE


                        '------------------------------------
                        'START: Thrust 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

             if_nz      mov    THRUST_level, #0

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, sfx_id_mask


                        '------------------------------------
                        'START: Mutant 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

             if_nz      mov     MUTANT_sample, MUTANT_SAMPLE_1
             if_nz      mov     MUTANT_sample_tick, #1
             if_nz      mov     MUTANT_sample_index, #MUTANT_SAMPLES_PER_LWORD
             if_nz      mov     MUTANT_sample_width, #MUTANT_SAMPLE_WIDTH_INIT

                        'Indicate that the sound is in progress
             if_nz      or     sfx_in_progress, sfx_id_mask


                        '------------------------------------
                        'START: Lander Fire 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

              if_nz     mov     LFIRE_sample, LFIRE_SAMPLE_1
              if_nz     mov     LFIRE_sample_tick, #1
              if_nz     mov     LFIRE_sample_index, #LFIRE_SAMPLES_PER_LWORD
              if_nz     mov     LFIRE_sample_width, #LFIRE_SAMPLE_WIDTH_INIT

                        'Indicate that the sound is in progress
              if_nz     or     sfx_in_progress, sfx_id_mask

              
                        '------------------------------------
                        'START: Lander Die 
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

              if_nz     mov     LDIE_angle, #0
              if_nz     mov     LDIE_angle_step, LDIE_ANGLE_STEP_INIT
              if_nz     mov     LDIE_scale, #4

                        'Indicate that the sound is in progress
              if_nz     or     sfx_in_progress, sfx_id_mask

                        '------------------------------------
                        'START: Human Abducted
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

               if_nz    mov     HABD_angle0, #0
               if_nz    mov     HABD_angle1, #0  
               if_nz    mov     HABD_angle2, #0
               if_nz    mov     HABD_state_counter, #0
               if_nz    mov     HABD_snd0_levels, HABD_SND0_LEVEL
               if_nz    mov     HABD_snd1_levels, HABD_SND1_LEVEL
               if_nz    mov     HABD_segment, #15

                        'Indicate that the sound is in progress
               if_nz    or     sfx_in_progress, sfx_id_mask

                        '------------------------------------
                        'START: Human Fall
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

               if_nz    mov     HFALL_angle, #0
               if_nz    mov     HFALL_angle_step, HFALL_ANGLE_STEP_INIT
               if_nz    mov     HFALL_state_counter, HFALL_DURATION

                        'Indicate that the sound is in progress
               if_nz    or     sfx_in_progress, sfx_id_mask
                                                                         
                        '------------------------------------
                        'START: Human Caught
                        '------------------------------------
                        shl     sfx_id_mask,#1
                        test    start_request_actions, sfx_id_mask    wz

               if_nz    mov     HCAUGHT_angle, #0
               if_nz    mov     HCAUGHT_angle_step, HCAUGHT_ANGLE_STEP_INIT
               if_nz    mov     HCAUGHT_state_counter, #0

                        'Indicate that the sound is in progress
               if_nz    or     sfx_in_progress, sfx_id_mask


                        '================================================================================
                        'PLAY: Explode
                        '================================================================================
                        add     EXPLODE_state_counter, #1 
 
                        mov     sfx_id_mask, #SOUND__INT_EXPLODE
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:explode_bypass

                        '------------------------------------
                        ' [You Die] Start the 2nd "bang" at the appropriate time
                        '------------------------------------
                        cmp     EXPLODE_type, #SOUND__YOU_DIE  wz
                  if_nz jmp     #:explode_you_die_bypass  
                  if_z  cmp     EXPLODE_state_counter, YOUDIE_BANG2_TIME wz
                  if_z  mov     EXPLODE_chunk_width, #2

                        '------------------------------------
                        ' [You Die] Process lingering rumble
                        '------------------------------------
                        cmp     EXPLODE_state_counter, YOUDIE_RUMBLE_TIME wc
                if_nc   mov     x, random_number
                if_nc   and     x, #$FF
                if_nc   add     x, #1        'keeps end case from being met prematurely
                if_nc   mov     EXPLODE_chunk_width, x
                if_nc   sub     EXPLODE_pulse_level, YOUDIE_PULSE_DECAY

                        '------------------------------------
                        ' [You Die] Process sound end
                        '------------------------------------
                        cmp     EXPLODE_state_counter, YOUDIE_END_TIME wz
                if_z    andn    sfx_in_progress, sfx_id_mask
                                  
:explode_you_die_bypass

                        '------------------------------------
                        ' [Planet Explode] Start the chaotic "bangs" at random times
                        '------------------------------------
                        cmp     EXPLODE_type, #SOUND__PLANET_EXPLODE  wz,wc
                  if_z  cmp     EXPLODE_state_counter, PLANET_CHAOS_TIME wc
           if_nz_or_nc  jmp     #:explode_skip0
                        mov     x, random_number
                        and     x, CHAOS_PROBABILITY_MASK
                        cmp     x, #1    wz
                  if_z  mov     EXPLODE_chunk_width, #2  

:explode_skip0
                        '------------------------------------
                        'Generate explosion sound
                        '------------------------------------
                        sub     EXPLODE_phase_ticks, #1 wc
                  if_nc jmp     #:explode_skip1
                        mov     EXPLODE_phase_ticks, FIRE_PHASE_TICK_INIT
                        add     EXPLODE_chunk_width, EXPLODE_chunk_step
                        cmp     EXPLODE_chunk_width, #60 wz     ' Reacned max 
                  if_nz cmp     EXPLODE_chunk_width, #0  wz     ' Reached min (for descending warp)     
                  if_z  andn    sfx_in_progress, sfx_id_mask 
:explode_skip1                            
                        sub     EXPLODE_pulse_ticks, #1  wz
                  if_nz jmp     #:explode_skip2
                        mov     x, random_number
                        and     x, #7
                        add     x, #1
                        mov     y, EXPLODE_chunk_width
                        call    #mult
                        mov     EXPLODE_pulse_ticks, y           
                        xor     EXPLODE_pulse_sense, #1

                        test    EXPLODE_pulse_sense, #1 wz    'Set edge flags to add rise/fall edge emphasis
                if_z    or      flags, #FLAG_FALLING_EDGE        
                if_nz   or      flags, #FLAG_RISING_EDGE           

:explode_skip2
                        'Output high if PULSE_BIT is high in either pulse generator
                        mov     AUDIO_channel_1, #00

                        test    EXPLODE_pulse_sense, #1 wz
                if_nz   mov     x, EXPLODE_pulse_level
                if_nz   shr     x, #22
                if_nz   mov     AUDIO_channel_1, x
          
:explode_bypass


                        '================================================================================
                        'PLAY: Thrust 
                        '================================================================================  
                        mov     sfx_id_mask, #SOUND__THRUST       
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:thrust_bypass

                        'Stop playback if requested
                '        test    stop_request_current, sfx_id_mask wz
                '  if_nz andn    sfx_in_progress, sfx_id_mask 

                  '     'Generate sound
                        mov     x, random_number
                        and     x, #$1fe
                        mov     AUDIO_channel_0, x
                        mov     y, EXPLODE_state_counter
                        and     y, #$F
                        cmp     y, #1 wz
                  if_nz mov     AUDIO_channel_0, THRUST_level        
                  if_z  mov     THRUST_level, x
                        
:thrust_bypass


                        '================================================================================
                        'PLAY: Mutant 
                        '================================================================================
:sfx_play_mutant 
                        shl     sfx_id_mask, #1
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:mutant_bypass

                        sub     MUTANT_sample_tick, #1   wz          
                  if_nz jmp     #:mutant_render

                        mov     MUTANT_sample_tick, MUTANT_sample_width         'Reset the tick
                        shr     MUTANT_sample_tick, #3
                        add     MUTANT_sample_tick, #1
                        
                        shr     MUTANT_sample, #2                               'Advance to next 2 bit sample
                        sub     MUTANT_sample_index, #1  wz
                  if_nz jmp     #:mutant_render

                        
                        mov     MUTANT_sample_index, #MUTANT_SAMPLES_PER_LWORD  'Reset the sample index

                        
                        mov     x, MUTANT_SAMPLE_1                              'Swap the mutant sample words 
                        mov     MUTANT_SAMPLE_1, MUTANT_SAMPLE_2
                        mov     MUTANT_SAMPLE_2, x

                        mov     MUTANT_sample, MUTANT_SAMPLE_1                  'Get the next sample word

                        add     MUTANT_sample_width, #1                         'Increase the sample width       

                        cmp     MUTANT_sample_width, #$100 wz
                  if_z  andn    sfx_in_progress, sfx_id_mask                    'Stop when max width reached

                  
                        '------------------------------------
                        'Render the 2 bit mutant sample
                        '------------------------------------    
:mutant_render          mov     AUDIO_channel_1, MUTANT_sample
                        and     AUDIO_channel_1, #3
                        shl     AUDIO_channel_1, #6
               
:mutant_bypass

                        '================================================================================
                        'PLAY: Lander Fire 
                        '================================================================================
:sfx_play_lfire

                        shl     sfx_id_mask, #1
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:lfire_bypass

                        sub     LFIRE_sample_tick, #1   wz          
                  if_nz jmp     #:lfire_render

                        mov     LFIRE_sample_tick, LFIRE_sample_width          'Reset the tick
                        shr     LFIRE_sample_tick, #2
                        add     LFIRE_sample_tick, #1
                        
                        shr     LFIRE_sample, #1                               'Advance to next 1 bit sample
                        sub     LFIRE_sample_index, #1  wz
                  if_nz jmp     #:LFIRE_render

                        
                        mov     LFIRE_sample_index, #LFIRE_SAMPLES_PER_LWORD   'Reset the sample index

                        
                        mov     x, LFIRE_SAMPLE_1                              'Swap the LFIRE sample words 
                        mov     LFIRE_SAMPLE_1, LFIRE_SAMPLE_2
                        mov     LFIRE_SAMPLE_2, x

                        mov     LFIRE_sample, LFIRE_SAMPLE_1                   'Get the next sample word

                        add     LFIRE_sample_width, #1                         'Increase the sample width       

                        cmp     LFIRE_sample_width, #$20 wz
                  if_z  andn    sfx_in_progress, sfx_id_mask                   'Stop when max width reached

                  
                        '------------------------------------
                        'Render the 2 bit LFIRE sample
                        '------------------------------------    
:lfire_render           mov     AUDIO_channel_1, LFIRE_sample
                        and     AUDIO_channel_1, #1
                        shl     AUDIO_channel_1, #8
                 
:lfire_bypass

                        '================================================================================
                        'PLAY: Lander Die 
                        '================================================================================  
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:ldie_bypass


                        mov     flags, #0   'EPMTEMP: TEMP FIX TO MASK EXPLODE EDGES DURING LANDER DIE ************************************
                        
                        '------------------------------------
                        'Compute waveform
                        '------------------------------------

                        'Get sine of current angle
                        mov     r1, LDIE_angle_step
                        shl     r1,#3
                        add     LDIE_angle, r1
                        mov     sin, LDIE_angle
                        shr     sin, #10
                        call    #getsin
                        add     sin, HALF_MAX_SIN
                        shr     sin, #9

                        'Scale by current scaling factor (4,3,2,1)
                        mov     x, sin
                        mov     y, LDIE_scale
                        call    #mult
                        mov     sin, y

                        'Clip
                        cmp     sin, #$1ff     wc
                 if_nc  mov     sin, #$1ff

                        'Decay the angle step
                        mov     x, LDIE_angle_step
                        mov     y, LDIE_ANGLE_MULTIPLY
                        call    #mult
                        shr     y, #15
                        mov     LDIE_angle_step, y

                        'When angle step has decayed update scaling factor
                        cmp     LDIE_angle_step, #$ff wz
                  if_nz jmp     #:ldie_skip1
                        mov     LDIE_angle, #0
                        mov     LDIE_angle_step, LDIE_ANGLE_STEP_INIT
                        sub     LDIE_scale, #1 wz
                  if_z  andn    sfx_in_progress, sfx_id_mask
                        
:ldie_skip1
                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        mov     AUDIO_channel_1, sin
:ldie_bypass

                        '================================================================================
                        'PLAY: Human abducted 
                        '================================================================================  
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:habd_bypass


                        mov     AUDIO_channel_1, #0
                        add     HABD_state_counter, #1

                        cmp     HABD_state_counter, HABD_SEGMENT_TIME wc
                 if_nc  mov     HABD_state_counter, #0
                 if_nc  mov     HABD_angle0, #0
                 if_nc  sub     HABD_segment, #1 wz
                 if_nc  shr     HABD_snd0_levels, #2
                 if_nc  shr     HABD_snd1_levels, #2 
           if_nc_and_z  andn    sfx_in_progress, sfx_id_mask

                        cmp     HABD_state_counter, HABD_SEGMENT_ON_TIME wc
                 if_nc  jmp     #:habd_bypass
                        
                        '------------------------------------
                        'Compute waveform
                        '------------------------------------

                        'Compute frequency swell
                        
                        add     HABD_angle0, HABD_SWELL_STEP 
                        mov     sin, HABD_angle0
                        shr     sin, #10
                        call    #getsin
                        sar     sin, #1
                        
                        'Get sine of current angle
                        add     HABD_angle1, sin                    'add in frequency swelling
                        add     HABD_angle1, HABD_ANGLE_STEP        'add in base frequency
                        mov     sin, HABD_angle1
                        shr     sin, #10
                        call    #getsin
                        mov     r1, sin 
                        shl     sin, #3
                        add     sin, HABD_ANGLE_STEP
                        

                        add     HABD_angle2, sin
                        mov     sin, HABD_angle2
                        shr     sin, #10
                        call    #getsin
                        add     sin, HALF_MAX_SIN
                        shr     sin, #7

                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        mov     x, HABD_snd0_levels
                        and     x, #$3
                        shr     sin, x
                        mov     AUDIO_channel_1, sin
                        
                        neg     r1, r1
                        add     r1,HALF_MAX_SIN
                        shr     r1, #9
                        cmp     r1, #$Fc wc
                        mov     r2, #0
                 if_nc  mov     r2, #$180 

                        mov     x, HABD_snd1_levels
                        and     x, #$3
                        shr     r2, x
                        add     AUDIO_channel_1, r2
                        
                 
                         
:habd_bypass
                        '================================================================================
                        'PLAY: Human Fall
                        '================================================================================  
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:hfall_bypass

                        
                        '------------------------------------
                        'Compute waveform
                        '------------------------------------

                        'Get sine of current angle
                        add     HFALL_angle, HFALL_angle_step
                        sub     HFALL_angle_step, #$3
                        mov     sin, HFALL_angle
                        shr     sin, #10
                        call    #getsin
                        add     sin, HALF_MAX_SIN
                        shr     sin, #9
                        and     sin, #$C0    ' Drop lower bits to give a little grainyness   

                        '------------------------------------
                        'Determine end time
                        '------------------------------------
                        sub     HFALL_state_counter, #1  wz
                  if_z  andn    sfx_in_progress, sfx_id_mask   

                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        mov     AUDIO_channel_1, sin
:hfall_bypass


                        '================================================================================
                        'PLAY: Human Caught
                        '================================================================================  
                        shl     sfx_id_mask, #1 
                        test    sfx_in_progress, sfx_id_mask  wz
                  if_z  jmp     #:hcaught_bypass

                        '------------------------------------
                        'Compute waveform
                        '------------------------------------

                        'Get sine of current angle
                        add     HCAUGHT_angle, HCAUGHT_angle_step
                        add     HCAUGHT_angle_step, #$20
                        mov     sin, HCAUGHT_angle
                        shr     sin, #10
                        call    #getsin
                        add     sin, HALF_MAX_SIN
                        shr     sin, #9
                        and     sin, #$C0    ' Drop lower bits to emulate "stair step" of original sound

                        '------------------------------------
                        'Determine end time
                        '------------------------------------
                        add     HCAUGHT_state_counter, #1
                        
                        cmp     HCAUGHT_state_counter, HCAUGHT_SWEEP2_TIME wz
                  if_z  mov     HCAUGHT_angle_step, HCAUGHT_ANGLE_STEP_INIT
                  
                        cmp     HCAUGHT_state_counter, HCAUGHT_SWEEP3_TIME wz
                  if_z  mov     HCAUGHT_angle_step, HCAUGHT_ANGLE_STEP_INIT
                  if_z  mov     HCAUGHT_step_increment, #20 

                        cmp     HCAUGHT_state_counter, HCAUGHT_SWEEP_FAST_TIME wc
                  if_nc add     HCAUGHT_step_increment, #11              
                  if_nc mov     r2, HCAUGHT_step_increment
                  if_nc shr     r2, #7 
                  if_nc add     HCAUGHT_angle_step, r2
                  
                        cmp     HCAUGHT_state_counter, HCAUGHT_END_TIME wz
                  if_z  andn    sfx_in_progress, sfx_id_mask   

                        '------------------------------------
                        'Set PWM output
                        '------------------------------------
                        mov     AUDIO_channel_1, sin
:hcaught_bypass

                        '================================================================================
                        'Audio Mixing
                        '================================================================================

                        '------------------------------------
                        'Mix audio channels
                        '------------------------------------
                        mov     r2, AUDIO_channel_0
                        add     r2, AUDIO_channel_1                                                     
                        shl     r2, #18
                        add     r2, LONG_HALF

                        '------------------------------------
                        'Add edge emphasis to increase square wave rise times
                        '------------------------------------
                        
                        'Add edge emphasis if transitioning
                        test    flags, #FLAG_RISING_EDGE   wz
                  if_nz add     r2, EDGE_EMPHASIS_INIT                          'add edge emphasis
                        test    flags, #FLAG_FALLING_EDGE  wz
                  if_nz sub     r2, EDGE_EMPHASIS_INIT                          'add edge emphasis
                        
                        '------------------------------------
                        'Output wave data
                        '------------------------------------
                        mov     frqa, r2

                        

                        '================================================================================
                        'Done main loop
                        '================================================================================
                        
                        jmp     #:sample_sync


'------------------------------------ 
' Get sine/cosine
'
'       quadrant:    1            2            3            4
'          angle:    $0000..$07FF $0800..$0FFF $1000..$17FF $1800..$1FFF
'    table index:    $0000..$07FF $0800..$0001 $0000..$07FF $0800..$0001
'         mirror:    +offset      -offset      +offset      -offset
'           flip:    +sample      +sample      -sample      -sample
'
' on entry: sin[12..0] holds angle (0 to just under 360)
' on exit: sin holds signed value ranging from $0000FFFF ('1') to
' $FFFF0001 ('-1')
'------------------------------------ 
getcos                  add     sin,sin_90      'for cosine, add 90
getsin                  test    sin,sin_90 wc   'get quadrant 2|4 into c
                        test    sin,sin_180 wz  'get quadrant 3|4 into nz
                        negc    sin,sin         'if quadrant 2|4, negate offset
                        or      sin,sin_table   'or in sin table address >> 1
                        shl     sin,#1          'shift left to get final word address
                        rdword  sin,sin         'read word sample from $E000 to $F000
                        negnz   sin,sin         'if quadrant 3|4, negate sample
getsin_ret
getcos_ret              ret                     '39..54 clocks
                                                '(variance due to HUB sync on RDWORD)
sin_90                  long    $0800
sin_180                 long    $1000
sin_table               long    $E000 >> 1      'sine table base shifted right
sin                     long    0
HALF_MAX_SIN            long    $FFFF                               


'------------------------------------
'Multiply                                    
'------------------------------------
' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
'------------------------------------
mult                    shl x,#16               'get multiplicand into x[31..16]
                        mov t,#16               'ready for 16 multiplier bits
                        shr y,#1 wc             'get initial multiplier bit into c
:loop
                        if_c add y,x wc         'conditionally add multiplicand into product
                        rcr y,#1 wc             'get next multiplier bit into c.
                                                ' while shift product
                        djnz t,#:loop           'loop until done
mult_ret                ret                     'return with product in y[31..0]

'------------------------------------
'Initialized Data                                      
'------------------------------------
_PIN__AUDIO             long    PIN__AUDIO

SAMPLE_INTERVAL_HYDRA   long   (CLOCK_RATE_HYDRA / AUDIO_SAMPLE_RATE)
SAMPLE_INTERVAL_HYBRID  long   (CLOCK_RATE_HYBRID / AUDIO_SAMPLE_RATE)    

CTRAVAL                 long   %00110 << 26 + 7  'DUTY single-ended, APIN=7        '
LONG_HALF               long   $7FFF_FFFF

LFSR_32_TAPS            long   $800000c2

EDGE_EMPHASIS_INIT      long   $0fffffff

LDIE_ANGLE_STEP_INIT    long   $0000693a
LDIE_ANGLE_MULTIPLY     long   $00007Fd9 

FIRE_PHASE_TICK_INIT    long   $00000300

RANDOM_DELTA            long   $9e3779b9

SGAME_ANGLE_STEP        long   $0000f587
SGAME_ANGLE_END         long   $00800000

YOUDIE_BANG2_TIME       long   $00001000
YOUDIE_RUMBLE_TIME      long   $00007000
YOUDIE_END_TIME         long   $00020000
YOUDIE_PULSE_DECAY      long   $00005000

PLANET_CHAOS_TIME       long   $0001C000
CHAOS_PROBABILITY_MASK  long   $000007FF

MUTANT_SAMPLE_1         long   %01_10_00_11_00_10_01_00___01_10_00_11_00_10_01_00  ' 2 bit sample representation
MUTANT_SAMPLE_2         long   %01_01_01_01_01_01_01_01___01_10_00_11_00_10_01_00  ' of the mutant waveform                                                                                 

LFIRE_SAMPLE_1          Long   %01111101_11110111_11010110_00100101  ' 1 Bit sample representation 
LFIRE_SAMPLE_2          long   %00000000_00000101_00100011_01011111  ' of the lander fire waveform

HABD_SWELL_STEP         long   $00001800  
HABD_ANGLE_STEP         long   $00040500

HABD_SEGMENT_TIME       long   $00000500
HABD_SEGMENT_ON_TIME    long   $00000480
                               'These are 2 bit representations of the sound levels of the two sound components
                               'in the HABD sound for each segment, left to right in time
HABD_SND0_LEVEL         long   %11_10_10_10_01_00_00_01_01_10_11_11_00_10_00_00  ' Wavey sound   
HABD_SND1_LEVEL         long   %11_00_01_00_00_00_00_00_01_01_10_00_11_11_11_11  ' Sharp pulse sound

HFALL_ANGLE_STEP_INIT   long   $00060500
HFALL_DURATION          long   $00018000

HCAUGHT_ANGLE_STEP_INIT long   99383
HCAUGHT_SWEEP2_TIME     long   3753   
HCAUGHT_SWEEP3_TIME     long   7506
HCAUGHT_SWEEP_FAST_TIME long   9500
HCAUGHT_END_TIME        long   12495

'------------------------------------
'Uninitialized Data
'------------------------------------
                          
r1                        res     1
r2                        res     1
sfx_id_mask               res     1
flags                     res     1
sample_interval_ticks     res     1

x                         res     1    ' for multiply
y                         res     1    ' for multiply
t                         res     1    ' for multiply

previous_pulse_level      res     1

start_request_previous    res     1
start_request_current     res     1
start_request_actions     res     1
sfx_in_progress           res     1
stop_request_current      res     1

p_sfx_start_flags         res     1    
p_sfx_stop_flags          res     1
sfx_playing_mask          res     1

next_sample_start_time    res     1

random_number             res     1

'EXPLODE parameters
EXPLODE_phase_ticks       res     1
EXPLODE_chunk_width       res     1
EXPLODE_chunk_step        res     1
EXPLODE_pulse_ticks       res     1
EXPLODE_pulse_sense       res     1
EXPLODE_state_counter     res     1
EXPLODE_type              res     1
EXPLODE_pulse_level       res     1  

'LADER DIE parameters
LDIE_angle                res     1
LDIE_angle_step           res     1
LDIE_scale                res     1

'THRUST parameters
THRUST_level              res     1

'Audio channels
AUDIO_channel_0           res     1  
AUDIO_channel_1           res     1    

'START parameters
SGAME_angle               res     1
SGAME_scale               res     1

'MUTANT parameters 
MUTANT_sample             res     1
MUTANT_sample_tick        res     1
MUTANT_sample_index       res     1
MUTANT_sample_width       res     1

'LANDER FIRE parameters 
LFIRE_sample              res     1
LFIRE_sample_tick         res     1
LFIRE_sample_index        res     1
LFIRE_sample_width        res     1

'HUMAN ABDUCTED parameters
HABD_state_counter        res     1
HABD_angle0               res     1
HABD_angle1               res     1
HABD_angle2               res     1
HABD_snd0_levels          res     1
HABD_snd1_levels          res     1
HABD_segment              res     1

'HUMAN FALL parameters
HFALL_angle               res     1
HFALL_angle_step          res     1
HFALL_state_counter       res     1

'HUMAN CAUGHT parameters
HCAUGHT_angle             res     1
HCAUGHT_angle_step        res     1
HCAUGHT_state_counter     res     1
HCAUGHT_step_increment    res     1

                          fit                 